#include	<stdio.h>
#include	<sys/types.h>
#include	<time.h>
#include	<sys/times.h>
#include	<sys/timeb.h>
#include	<signal.h>
#include	<utmp.h>

#define		VERSION		"1.1"

#define		VERBOSE		0x0001
#define		FORK		0x0002
#define		QUIET		0x0004
#define		DEBUG		0x0010

#define		PROTFILE	"/usr/adm/messages"
#define		CYCLETIME	600

long		cycletime = CYCLETIME;
long		mode = 0;
char		*protfile = PROTFILE;

extern struct tm *localtime();
extern struct utmp *getutent();

main(argc,argv)
int	argc;
char	**argv;
{
	int		erg;
	FILE		*pfp;
	FILE		*wfp;
	struct	utmp	*utmp;
	struct	tm	*tm;
	long		now;
	long		i,j;
	extern char	*optarg;
	extern int	optind;

	while((erg = getopt(argc,argv,"vf:bDht:")) != EOF) {
		switch(erg) {
		case 'h':
		default:
		case '?':
			fprintf(stderr,"portwd V%s\nusage: %s [-v] [-b] [-f file] [-t cycletime] port [port ..]\n",
				VERSION, argv[0]);
			fprintf(stderr,"	-f file	file for messages (def.: %s)\n",protfile);
			fprintf(stderr,"	-v	verbose mode (a lot of messages)\n");
			fprintf(stderr,"	-b	run as a background job\n");
			fprintf(stderr,"	-t n	cycle time to watch ports [sec] (def.: %d)\n",cycletime);
			fprintf(stderr,"	port	port to watch\n");
			exit(1);
		case 'D':
			mode |= DEBUG;
			break;
		case 'f':
			protfile = optarg;
			break;
		case 'b':
			mode |= FORK;
			break;
		case 'v':
			mode |= VERBOSE;
			break;
		case 't':
			cycletime = atoi(optarg);
			break;
		}
	}

	if (optind >= argc) {
		fprintf(stderr,"you should watch at least one port\n");
		exit(1);
	}
	if ((pfp = fopen(protfile,"a")) == NULL) {
		fprintf(stderr,"Cannot open %s\n",protfile);
		exit(1);
	}

	now = time((long)0);
	tm = localtime(&now);
	fprintf(pfp,"%s%s started watching ",asctime(tm),argv[0]);
	for(i=optind; i<argc; i++) fprintf(pfp,"%s ",argv[i]);
	fprintf(pfp,"\n");
	if (pfp) fclose(pfp);

	if (mode & FORK) {
		if (fork()) return(0);
		if (!setpgrp()) {
			fprintf(stderr,"Cannot set new process group\n");
			exit(1);
		}
		signal(SIGINT, SIG_IGN);
		signal(SIGHUP, SIG_IGN);
	}

	while(1) {
		if ((pfp = fopen(protfile,"a")) == NULL) {
			fprintf(stderr,"Cannot open %s\n",protfile);
			exit(1);
		}
		now = time((long)0);
		tm = localtime(&now);
		if (mode & VERBOSE) {
			fprintf(pfp,"%s%s watches\n",asctime(tm),argv[0]);
		}
		while(utmp = getutent()) {
			for(i=optind; i<argc; i++) {
				if(!strncmp(utmp->ut_line, argv[i], sizeof(utmp->ut_line))) {
					if (mode & VERBOSE) {
						fprintf(pfp,"	%-8.8s %-4.4s %-12.12s %05d %01d %02d/%02d %05d\n",
							utmp->ut_user, utmp->ut_id, utmp->ut_line,
							utmp->ut_pid, utmp->ut_type, utmp->ut_exit.e_exit,
							utmp->ut_exit.e_termination, utmp->ut_time
						);
					}
					if (!strncmp(utmp->ut_user,"LOGIN",sizeof(utmp->ut_name))) {
						/* port ist auf jeden fall frei */
						j = trykill(utmp->ut_pid);
						if (j) {
							fprintf(pfp,"%s%s: getty on port %-.12s (pid=%d) successfully killed (signal=%d)\n",asctime(tm),argv[0],utmp->ut_line,utmp->ut_pid,j);
						} else {
							fprintf(pfp,"%s%s: couldn't kill getty on port %-.12s (pid=%d)\n",asctime(tm),argv[0],utmp->ut_line,utmp->ut_pid);
						}
					}
				}
			}
		}
		fclose(pfp);
		endutent();
		sleep(cycletime);
	}
}

int
trykill(pid)
long	pid;
{
	if (kill(pid,SIGTERM)) {
		if (kill(pid,SIGKILL)) {
			return(0);
		} else {
			return(SIGKILL);
		}
	} else {
		return(SIGTERM);
	}
}
